from time import time

import shmem_utils
from karabo.bound import KARABO_CLASSINFO
from PipeToZeroMQ import PipeToZeroMQ, conversion

from ._version import version as deviceVersion


@KARABO_CLASSINFO("ShmemToZMQ", deviceVersion)
class ShmemToZMQ(PipeToZeroMQ.PipeToZeroMQ):
    @staticmethod
    def expectedParameters(expected):
        super(ShmemToZMQ, ShmemToZMQ).expectedParameters(expected)

    def initialization(self):
        """Just overriding to avoid KARABO_ON_INPUT"""

        self.KARABO_SLOT(self.activate)
        self.KARABO_SLOT(self.deactivate)
        self.KARABO_SLOT(self.reset)
        self.KARABO_SLOT(self.requestScene)
        # ON_DATA is a bit more convenient I think
        self.KARABO_ON_DATA("input", self.on_data)
        self._source_to_shmem_ary = {}
        self._source_to_shmem_mem = {}

    def on_data(self, data, metadata):
        source = metadata["source"]
        if source not in self.sources:
            self.log.INFO(f"New direct source encountered: {source}")
            schema = self.getFullSchema()
            self.appendSchema(PipeToZeroMQ.timestamp_schema(schema, source))
            self.sources.add(source)

        actual = self.getActualTimestamp()
        input_tic = time()
        self.info["inputUpdated"] += 1
        self._time_info(metadata, actual, self.info)

        new_meta = {}
        new_data = {}

        # additional unstacked sources
        if data.has("unstacked"):
            for source in data.get("unstacked").getKeys():
                forward, ignore = self._filter_properties(data, source)
                new_data[source] = conversion.hash_to_dict(
                    data.get(f"unstacked.{source}"), paths=forward, version=self.version
                )
            data.erase("unstacked")

        # stacked data from ModuleStacker
        for source, handle_string in zip(data.get("sources"), data.get("image.data")):
            if source not in self.sources:
                self.log.INFO(f"New original source encountered: {source}")
                self.sources.add(source)

            if self.allowed_sources and source not in self.allowed_sources:
                continue

            forward, ignore = self._filter_properties(data, source)

            new_meta[source] = conversion.meta_to_dict(metadata)
            new_meta[source]["ignored_keys"] = ignore

            # may have to open shared memory buffer
            if source not in self._source_to_shmem_ary:
                self.log.INFO(
                    f"For source {source}, will open shmem handle {handle_string}"
                )
                mem, ary = shmem_utils.open_shmem_from_handle(handle_string, rw=False)
                self._source_to_shmem_mem[source] = mem
                self._source_to_shmem_ary[source] = ary

            # grab data from shared memory buffer
            _, _, _, index = shmem_utils.parse_shmem_handle(handle_string)
            buffer_data = self._source_to_shmem_ary[source][index]

            # conversion separates array stuf into separate dict, "arr"
            dic, arr = conversion.hash_to_dict(
                data, paths=forward, version=self.version
            )

            # and image.data was list of strings (in dic), to be replaced
            del dic["image.data"]
            arr["image.data"] = buffer_data
            new_data[source] = (dic, arr)
            # note: assumes version "2.2" of the PipeToZeroMQ packing format

        # forward data to all connected ZMQ sockets
        self._send(new_data, new_meta)

        output_tic = time()
        self.info["onInputTotal"] = 1000 * (output_tic - input_tic)

        # update properties on device
        self._updateProperties(output_tic)
        # block if device is in passive state
        self.monitoring.wait()
