import threading
import time
import timeit

import numpy as np

numpy_dtype_to_c_type_str = {
    np.uint16: "unsigned short",
    np.uint32: "unsigned short",
    np.float16: "half",  # warning: only in CUDA with special support
    np.float32: "float",
    np.float64: "double",
}


def ceil_div(num, denom):
    return (num + denom - 1) // denom


class DelayableTimer:
    """Start a timer which can be extended

    Useful for reverting to state after inactivity, for instance.

    timer defaults to timeit.default_timer - it should be a timer returning
    globally increasing number of seconds.
    """

    def __init__(self, timeout, callback, timer=timeit.default_timer):
        self.timer = timer
        self.stop_time = self.timer() + timeout

        def runner():
            now = self.timer()
            while now < self.stop_time:
                diff = self.stop_time - now
                time.sleep(diff)
                now = self.timer()
            callback()

        self.thread = threading.Thread(target=runner)
        self.thread.start()

    def set_timeout(self, timeout):
        """Delay stop time to now + timeout

        If now + timeout is sooner than already set timeout, this does nothing"""
        self.stop_time = self.timer() + timeout

    def add_timeout(self, timeout):
        """Simply add timeout to current stop time"""
        self.stop_time += timeout


class RepeatingTimer:
    """Similar to DelayableTimer, but will keep running with pre-set intervals"""

    def __init__(self, interval, callback, timer=timeit.default_timer, start_now=True):
        self.timer = timer
        self.stopped = True
        self.interval = interval
        self.callback = callback
        if start_now:
            self.start()

    def delay(self):
        self.stop_time = self.timer() + self.interval

    def start(self):
        self.stopped = False
        self.stop_time = self.timer() + self.interval

        def runner():
            while not self.stopped:
                now = self.timer()
                while now < self.stop_time:
                    diff = self.stop_time - now
                    time.sleep(diff)
                    if self.stopped:
                        return
                    now = self.timer()
                self.callback()
                self.stop_time = self.timer() + self.interval

        self.thread = threading.Thread(target=runner)
        self.thread.start()

    def stop(self):
        self.stopped = True


class Throttler:
    """Similar to DelayableTimer, but will keep running with pre-set intervals"""

    def __init__(self, interval, timer=timeit.default_timer):
        self.timer = timer
        self.interval = interval
        self.latest_call = None

    def ready(self):
        if self.latest_call is None:
            return True
        else:
            return self.latest_call + self.interval <= self.timer()

    def update(self):
        self.latest_call = self.timer()

    def ready_update(self):
        time = self.timer()
        if self.latest_call is None or self.latest_call + self.interval <= time:
            self.latest_call = time
            return True
        else:
            return False
